// mobjToTable: copies over relevant properties for carrying from b to a
// a: the mobj or table to copy values into
// b: the mobj or table to copy values from
freeslot(
	"SKINCOLOR_DREAMY",
	"sfx_m2jump",
	"sfx_pluck",
	"sfx_toss",
	"sfx_m2ouch",
	"sfx_m2die",
	"sfx_m2life",
	"SPR_VEGI",
	"S_PLAY_PLUCK1",
	"S_PLAY_PLUCK2",
	"S_PLAY_CARRYCROUCH",
	"S_CARRYDISPLAY",
	"S_THROWNENEMY_SPAWN",
	"S_VEGGIE",
	"MT_CARRYDISPLAY",
	"MT_THROWNENEMY",
	"MT_VEGGIE"
)

local mobjToTable = function(a,b)
	if not (a and b
	and b.valid)
		return false end
		
	a.state = b.state
	a.sprite = b.sprite
	a.frame = b.frame
	a.anim_duration = b.anim_duration
	a.renderflags = b.renderflags
	a.blendmode = b.blendmode
	a.spritexscale = b.spritexscale
	a.spriteyscale = b.spriteyscale
	a.spritexoffset = b.spritexoffset
	a.spriteyoffset = b.spriteyoffset
	a.radius = b.radius
	a.height = b.height
	a.tics = b.tics
	a.flags = b.flags
	a.flags2 = b.flags2
	a.color = b.color
	a.colorized = b.colorized
	a.drawonlyforplayer = b.drawonlyforplayer
	a.dontdrawforviewmobj = b.dontdrawforviewmobj
	a.hnext = b.hnext
	a.hprev = b.hprev
	a.health = b.health
	a.movedir = b.movedir
	a.movecount = b.movecount
	a.target = b.target
	a.reactiontime = b.reactiontime
	a.threshold = b.threshold
	a.lastlook = b.lastlook
	a.spawnpoint = b.spawnpoint
	a.tracer = b.tracer
	a.friction = b.friction
	a.movefactor = b.movefactor
	a.fuse = b.fuse
	a.scale = b.scale
	a.destscale = b.destscale
	a.scalespeed = b.scalespeed
	a.extravalue1 = b.extravalue1
	a.extravalue2 = b.extravalue2
	a.cusval = b.cusval
	a.cvmem = b.cvmem
	a.mirrored = b.mirrored
	a.shadowscale = b.shadowscale
	a.dispoffset = b.dispoffset
	
	if b.hitfunction
		a.hitfunction = b.hitfunction
	end
	
	return true
end

local throwEnemy = function(mo,speed)
	if not mo and mo.valid
		return end
	
	if not (mo.carrying and mo.carrying.valid
	and mo.carrydisplay and mo.carrydisplay.valid)
		mo.carrying = nil
		mo.carrydisplay = nil
		return end
		
	S_StartSound(mo,sfx_toss)
	
	local thrown = P_SpawnMobj(mo.carrydisplay.x,mo.carrydisplay.y,mo.carrydisplay.z,MT_THROWNENEMY)
	if not thrown and thrown.valid
		return true end
			
	thrown.target = mo
	thrown.angle = mo.angle
	P_InstaThrust(thrown,mo.angle,speed)
	
	thrown.momx = $+mo.momx
	thrown.momy = $+mo.momy
	thrown.momz = $+mo.momz
	
	thrown.carrying = {}
	mobjToTable(thrown.carrying,mo.carrying)
	
	thrown.eflags = mo.eflags
	
	thrown.carrying.type = mo.carrying.type
	thrown.carrying.valid = true
	mo.carrying = nil
	
	if thrown.carrying.type == MT_NULL
		thrown.momx = $*2
		thrown.momy = $*2
	end
	
	return thrown
end

local A_SpawnCarriedEnemy = function(t,a,b)
	local enemy = P_SpawnMobj(t.x,t.y,t.z,t.carrying.type)
	if not (enemy and enemy.valid)
		return
	end

	mobjToTable(enemy,t.carrying)
	enemy.eflags = t.eflags
	
	if (enemy.landfunction
	and enemy.landfunction(enemy,t,t.hitsomething))
		return end
	
	if ~enemy.flags & MF_NOGRAVITY
		enemy.momx = t.momx
		enemy.momy = t.momy
		enemy.momz = t.momz / -2
		enemy.angle = t.angle
	end
	
	if t.hitsomething
		P_DamageMobj(enemy,t.target,t.target)
	end
end

skincolors[SKINCOLOR_DREAMY] = {
	name = "Dreamy",
	ramp = {
		0,
		144,
		145,
		147,
		148,
		149,
		150,
		151,
		152,
		153,
		154,
		155,
		155,
		156,
		157,
		158
	},
	invcolor = SKINCOLOR_ISLAND,
	invshade = 5,
	chatcolor = V_REDMAP,
	accessible = true
}

states[S_PLAY_PLUCK1] = {
	sprite = SPR_PLAY,
	frame = SPR2_MLEE,
	tics = TICRATE/6,
	nextstate = S_PLAY_PLUCK2
}

states[S_PLAY_PLUCK2] = {
	sprite = SPR_PLAY,
	frame = SPR2_LAND,
	tics = TICRATE/6,
	nextstate = S_PLAY_FALL
}

states[S_PLAY_CARRYCROUCH] = {
	sprite = SPR_PLAY,
	frame = SPR2_LAND,
	tics = -1,
	nextstate = S_PLAY_FALL
}

states[S_CARRYDISPLAY] = {
	sprite = SPR_NULL,
	frame = A,
	tics = -1
}

states[S_THROWNENEMY_SPAWN] = {
	sprite = SPR_NULL,
	frame = A,
	action = A_SpawnCarriedEnemy,
	tics = 0
}

states[S_VEGGIE] = {
	sprite = SPR_VEGI,
	frame = A,
	tics = 1,
	nextstate = S_SMOKE1
}

mobjinfo[MT_CARRYDISPLAY] = {
	spawnstate = S_CARRYDISPLAY,
	spawnhealth = 1000,
	painstate = S_CARRYDISPLAY,
	flags = MF_NOBLOCKMAP|MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOCLIPTHING
}

mobjinfo[MT_THROWNENEMY] = {
	spawnstate = S_CARRYDISPLAY,
	spawnhealth = 1,
	deathstate = S_THROWNENEMY_SPAWN,
	xdeathstate = S_THROWNENEMY_SPAWN,
	flags = MF_MISSILE
}

local veggie = {
	state = S_VEGGIE,
	sprite = SPR_VEGI,
	frame = A,
	anim_duration = 0,
	renderflags = 0,
	blendmode = AST_TRANSLUCENT,
	spritexscale = FU,
	spriteyscale = FU,
	spritexoffset = 0,
	spriteyoffset = 0,
	radius = 16*FU,
	height = 32*FU,
	tics = 1,
	flags = 0,
	flags2 = 0,
	color = 0,
	colorized = false,
	drawonlyforplayer = nil,
	dontdrawforviewmobj = nil,
	hnext = nil,
	hprev = nil,
	health = 0,
	movedir = 0,
	movecount = 0,
	target = nil,
	reactiontime = 0,
	threshold = 0,
	lastlook = 0,
	spawnpoint = nil,
	tracer = nil,
	friction = 0,
	movefactor = 0,
	fuse = 0,
	scale = FU,
	destscale = FU,
	scalespeed = FU/12,
	extravalue1 = 0,
	extravalue2 = 0,
	cusval = 0,
	cvmem = 0,
	mirrored = false,
	shadowscale = 0,
	dispoffset = 0,
	
	hitfunction = function(i,t)
		P_KillMobj(i)
		return true
	end
}

addHook("ShouldDamage",function(t,i,s,d,dt)
	if not (
		t and t.valid
		and i and i.valid
		and t.skin == "smb2"
	)
	or i.unstandable
	or i.player
	or (s and s.valid and s.player)
		return end
		
	if i.flags & MF_ENEMY
	and t.z - t.momz >= (i.z + i.height - i.momz) - FU
		i.madesolid = i.flags
		i.madesolidtime = 2
		i.flags = ($|MF_SOLID)&~(MF_SPECIAL|MF_FLOAT)
		i.madesolidtarget = t
		i.madesolidlastpos = {
			x = i.x,
			y = i.y,
			z = i.z
		}
		
		if i.madesolid & MF_NOGRAVITY
			P_TryMove(t,i.x,i.y,i.z+i.height,true)
			t.momz = 0
		end
		return false
	end
end,MT_PLAYER)

addHook("PostThinkFrame",function()
	for p in players.iterate
		p.carrylastbuttons = p.cmd.buttons
		
		if p.mo and p.mo.valid
			if (p.mo.state == S_PLAY_CARRYCROUCH
			or p.mo.state == S_PLAY_PLUCK1
			or p.mo.state == S_PLAY_PLUCK2)
				if p.mo.eflags & MFE_VERTICALFLIP
					p.mo.forthesakeofargumentiamsmall = 1
				else
					p.mo.height = P_GetPlayerSpinHeight(p)
				end
			elseif not P_PlayerShouldUseSpinHeight(p)
				p.mo.forthesakeofargumentiamsmall = 0
				p.mo.height = P_GetPlayerHeight(p)
			end
				
			if p.mo.jumpcharge
			and p.mo.jumpcharge > TICRATE/2
				p.mo.colorized = not $
				
				if p.mo.colorized
					p.mo.blendmode = AST_ADD
					p.mo.renderflags = RF_FULLBRIGHT
				else
					p.mo.blendmode = AST_TRANSLUCENT
					p.mo.renderflags = 0
				end
			elseif p.mo.blendmode == AST_ADD
				p.mo.colorized = false
				p.mo.color = p.skincolor
				p.mo.blendmode = AST_TRANSLUCENT
				p.mo.renderflags = 0
			end
			
			if (p.mo.sprite2 == SPR2_STND
			or p.mo.sprite2 == SPR2_WALK)
			and p.mo.carrying and p.mo.carrying.valid
				p.mo.sprite2 = SPR2_SKID
			end
		end
	end
	
	for mo in mobjs.iterate()
		if mo.type == MT_CARRYDISPLAY
			local t = mo.target
			if t and t.valid
			and t.carrying and t.carrying.valid
			and t.carrydisplay and t.carrydisplay == mo
				mo.dispoffset = t.dispoffset-1
				local z = t.height
				if t.forthesakeofargumentiamsmall
					z = P_GetPlayerSpinHeight(t.player)
				end
				
				if t.state == S_PLAY_PLUCK1
					z = ease.linear(t.tics*FU/(TICRATE/12),$,0)
					mo.dispoffset = t.dispoffset+1
				end
				
				mo.eflags = t.eflags
				if mo.eflags & MFE_VERTICALFLIP
					z = (-$)+t.height
				end
					
				P_MoveOrigin(mo,t.x,t.y,t.z+z)
			end
		
		elseif mo.flags & (MF_PUSHABLE|MF_MONITOR)
		and not mo.madesolid
		and not mo.unstandable
			local nearest = nil
			
			searchBlockmap("objects",function(r,f)
				if f.player
				and f.skin == "smb2"
				and R_PointToDist2(r.x,r.y,f.x,f.y) <= r.radius + f.radius
				and f.z <= r.z + r.height + FU
				and f.z > r.z
				and not (nearest and nearest.valid
					and R_PointToDist2(r.x,r.y,nearest.x,nearest.y) <= R_PointToDist2(r.x,r.y,f.x,f.y)
				)
					nearest = f
				end
			end,mo)
			
			if nearest
				mo.madesolid = mo.flags
				mo.madesolidtime = 2
				mo.flags = ($|MF_SOLID)&~(MF_SPECIAL|MF_FLOAT)
				mo.madesolidtarget = nearest
				mo.madesolidlastpos = {
					x = mo.x,
					y = mo.y,
					z = mo.z
				}
				
				mo.momz = 0
				if mo.madesolid & MF_NOGRAVITY
					P_TryMove(nearest,mo.x,mo.y,mo.z+mo.height,true)
					nearest.momz = 0
				end
			end
		end
		
		if mo.madesolidtime
		and mo.madesolidtarget
		and mo.madesolidtarget.skin == "smb2"
			local t = mo.madesolidtarget
			if R_PointToDist2(mo.x,mo.y,t.x,t.y) <= mo.radius + t.radius
			and t.z <= mo.z + mo.height + (32*FU)
			and t.z > mo.z
				t.standingon = mo
				mo.madesolidtime = 2
				
				if mo.madesolid & MF_NOGRAVITY
					mo.momz = min($,-FU/2)
					if not (t.player.pflags & PF_JUMPED)
						t.momz = min($,-FU/2)
					end
				end
				
				local delta = {
					x = mo.x - mo.madesolidlastpos.x,
					y = mo.y - mo.madesolidlastpos.y,
					z = mo.z - mo.madesolidlastpos.z
				}
				
				if delta.z > 0
					mo.z = $-delta.z
				end
				
				P_TryMove(t,t.x+delta.x,t.y+delta.y,mo.z+mo.height,true)
				
				if t.standingon and t.standingon.valid and t.standingon == mo
					P_SpawnLockOn(t.player,mo,S_LOCKON1)
				end
				
				mo.madesolidlastpos = {
					x = mo.x,
					y = mo.y,
					z = mo.z
				}
			else
				mo.madesolidtime = $-1
				t.flags = $&~MF_NOGRAVITY
			end
		elseif mo.madesolid
			mo.flags = mo.madesolid
			mo.madesolidtime = 0
			mo.madesolid = nil
			mo.madesolidtarget = nil
		end
	end
end)

addHook("PreThinkFrame",function()
	for p in players.iterate
		local mo = p.mo
		if not (mo and mo.valid)
			return end
			
		// dear whoever decided that PF_SLIDING should be cleared before the think frame on every frame:
		// millions are suffering every day because of your impotence.
		// there is blood on your hands. innocent blood. you monster.
		// slash j
		// seriously though PLEASE fix that
		mo.yetanotherslidingcheck = p.pflags & PF_SLIDING
			
		if P_IsObjectOnGround(mo)
			p.powers[pw_strong] = 0
			if p.cmd.buttons & BT_CUSTOM1
			and not (P_PlayerInPain(p)
			or p.pflags & (PF_SPINNING|PF_SLIDING|PF_STASIS)
			or p.powers[pw_carry]
			or p.playerstate == PST_DEAD)
				mo.state = S_PLAY_CARRYCROUCH
				p.cmd.forwardmove = 0
				p.cmd.sidemove = 0
				if p.speed > 0
					p.skidnow = 1
				end
				
				if mo.jumpcharge
					if mo.jumpcharge >= TICRATE/2
						mo.jumpcharge = TICRATE*3
					else
						mo.jumpcharge = $+1
					end
				else
					mo.jumpcharge = 1
				end
			else
				if mo.jumpcharge
					mo.jumpcharge = $-1
				else
					mo.jumpcharge = 0
				end
				
				if mo.state == S_PLAY_CARRYCROUCH
					mo.state = S_PLAY_STND
				end
			end
		else
			if mo.jumpcharge
			and mo.jumpcharge > TICRATE/2
			and p.pflags & PF_JUMPED
				mo.momz = $*2
				p.powers[pw_strong] = STR_BUST
			end
			
			mo.jumpcharge = 0
		end
	end
end)

addHook("SpinSpecial",function(p)
	if not (p.mo and p.mo.valid and p.mo.skin == "smb2") 
		return end
	
	if not (p.carrylastbuttons != nil
	and ~p.carrylastbuttons & BT_SPIN)
	or p.exiting
	or p.powers[pw_nocontrol]
	or p.powers[pw_carry]
	or p.mo.yetanotherslidingcheck
		return end
		
	local mo = p.mo
	
	if mo.carrying and mo.carrying.valid
		mo.state = S_PLAY_STND
		p.drawangle = mo.angle
		if p.prevdrawangle != nil
			p.prevdrawangle = mo.angle
		end
		local thrown = throwEnemy(mo,p.actionspd)
		if thrown.valid
		and thrown.eflags & MFE_VERTICALFLIP
			thrown.flags2 = $|MF2_OBJECTFLIP
			P_MoveOrigin(thrown,thrown.x,thrown.y,thrown.z - mo.height)
		end
		return true
	elseif mo.standingon
	and mo.standingon.valid
	and mo.standingon.madesolidtarget
	and mo.standingon.madesolidtarget == mo
	and mo.health
		local t = mo.standingon
		p.powers[pw_nocontrol] = TICRATE/3
		S_StartSound(mo,sfx_pluck)
		mo.state = S_PLAY_PLUCK1
		mo.carrying = {}
		
		mo.momx = $/2
		mo.momy = $/2
		mo.momz = $/2
		
		if not (t.pulledfunction
		and t.pulledfunction(t,mo,mo.carrying))
			mobjToTable(mo.carrying,t)
			
			mo.carrying.type = t.type
			
			if t.madesolid
				mo.carrying.flags = t.madesolid
			end
			
			mo.carrying.valid = true
			t.state = S_NULL
		end
		return true
	elseif P_IsObjectOnGround(mo)
	and mo.floorrover
	and (mo.floorrover.busttype == BT_STRONG
	or mo.floorrover.flags & FOF_BUSTUP)
		EV_CrumbleChain(mo.floorrover)
		p.powers[pw_nocontrol] = TICRATE/3
		S_StartSound(mo,sfx_pluck)
		mo.state = S_PLAY_PLUCK1
	else
		S_StartSound(mo,sfx_pluck)
		mo.state = S_PLAY_PLUCK1
		mo.momx = $/2
		mo.momy = $/2
		mo.momz = $/2
		
		mo.carrying = veggie
		mo.carrying.type = MT_NULL
		mo.carrying.valid = true
		mo.carrying.color = mo.color
	end
end)

addHook("PlayerThink",function(p)
	local mo = p.mo
	if not (mo and mo.valid)
		return end
		
	if mo.skin != "smb2"
		if mo.carrying and mo.carrying.valid
			throwEnemy(mo,skins["smb2"].actionspd)
		end
		
		if mo.jumpcharge
			mo.jumpcharge = nil
			mo.colorized = false
		end
		
		mo.carrying = nil
		mo.carrydisplay = nil
		mo.deathstilltimer = nil
		
	return end
	
	if mo.deathstilltimer
		if mo.momz > 0
		or mo.momz < -FU
			mo.storedmomz = mo.momz
		end
		mo.deathstilltimer = $-1
		P_SetObjectMomZ(mo, FixedMul(gravity,mo.subsector.sector.gravity), false)
	elseif mo.storedmomz
		mo.momz = mo.storedmomz
		mo.storedmomz = 0
	end
	
	if mo.carrying
	and (P_PlayerInPain(p)
	or p.pflags & (PF_SPINNING|PF_STASIS)
	or mo.yetanotherslidingcheck
	or p.powers[pw_carry]
	or p.playerstate == PST_DEAD)
		throwEnemy(mo,p.actionspd)
	end
	
	if mo.carrying
	and mo.carrying.valid
		if mo.carrydisplay
		and mo.carrydisplay.valid
		and mo.carrydisplay.type == MT_CARRYDISPLAY
			local c = mo.carrydisplay
			if mo.carrying.sprite
				c.sprite = mo.carrying.sprite
				c.frame = mo.carrying.frame & FF_FRAMEMASK
				c.renderflags = mo.carrying.renderflags
				c.color = mo.carrying.color
				c.colorized = mo.carrying.colorized
			end
			c.angle = p.drawangle
		else
			local c = P_SpawnMobj(mo.x,mo.y,mo.z+mo.height,MT_CARRYDISPLAY)
			mo.carrydisplay = c
			c.target = mo
		end
		
		if mo.carrying.type == MT_NULL
			p.normalspeed = 32*FU
		else
			p.normalspeed = 40*FU
		end
		p.runspeed = 1
	else
		p.normalspeed = 28*FU
		p.runspeed = 999*FU
	end
end)

addHook("MobjThinker",function(mo)
	if not (mo and mo.valid)
		return end
		
	local t = mo.target
	if not (t and t.valid
	and t.carrying and t.carrying.valid
	and t.carrydisplay and t.carrydisplay == mo)
		mo.state = S_NULL
	return end
end,MT_CARRYDISPLAY)

addHook("MobjThinker",function(mo)
	if not (mo and mo.valid)
		return end
		
	if not (mo.carrying and mo.carrying.valid)
		mo.state = S_NULL
	return end
	
	mo.sprite = mo.carrying.sprite
	mo.frame = mo.carrying.frame & FF_FRAMEMASK
	mo.renderflags = mo.carrying.renderflags
	mo.color = mo.carrying.color
	mo.colorized = mo.carrying.colorized
	mo.height = mo.carrying.height
	if mo.valid
		mo.radius = mo.carrying.radius
	end
end,MT_THROWNENEMY)

addHook("MobjMoveCollide",function(mo,t)
	if (t.type == MT_SPIKE
	or t.type == MT_WALLSPIKE
	or t.type == MT_WALLSPIKEBASE)
	and mo.z <= t.z + t.height
	and t.z <= mo.z + mo.height
		P_KillMobj(t)
		return false
	end
end,MT_THROWNENEMY)

addHook("ShouldDamage",function(t,i,s,d,dt)
	if (t and t.valid and t.player)
	and (gametyperules & GTR_FRIENDLY
	or ((s and s.valid and s.player)
	and s.player.ctfteam and t.player.ctfteam
	and s.player.ctfteam == t.player.ctfteam))
		return end
	
	if i and i.valid
	and i.type == MT_THROWNENEMY
		S_StartSound(t,sfx_m2ouch)
		if i.carrying.hitfunction
		and i.carrying.hitfunction(i,t)
			return end
			
		local thrown = P_SpawnMobj(i.x,i.y,i.z,MT_THROWNENEMY)
		if not thrown and thrown.valid
			return end
				
		thrown.target = i.target
		thrown.angle = i.angle
		thrown.hitsomething = true
		
		thrown.momx = i.momx/2
		thrown.momy = i.momy/2
		thrown.momz = -i.momz
		
		thrown.carrying = {}
		mobjToTable(thrown.carrying,i.carrying)
		
		thrown.carrying.type = i.carrying.type
		thrown.carrying.valid = true
		
		i.state = S_NULL
	end
end)

local nostand = function(mo) mo.unstandable = true end

addHook("MobjSpawn",nostand,MT_GFZFISH)
addHook("MobjSpawn",nostand,MT_SPINCUSHION)
addHook("MobjSpawn",nostand,MT_MINUS)
addHook("MobjSpawn",nostand,MT_GARGOYLE)
addHook("MobjSpawn",nostand,MT_MINECART)

local killwhenland = function(mo)
	
	mo.landfunction = function(t,i,hit)
		P_KillMobj(t)
		return true
	end
end

addHook("MobjSpawn",killwhenland,MT_JETJAW)
addHook("MobjSpawn",killwhenland,MT_CRUSHSTACEAN)
addHook("MobjSpawn",killwhenland,MT_SKIM)
addHook("MobjSpawn",killwhenland,MT_SNAILER)
addHook("MobjSpawn",killwhenland,MT_PTERABYTE)
addHook("MobjSpawn",killwhenland,MT_POINTY)
addHook("MobjSpawn",killwhenland,MT_DRAGONBOMBER)
addHook("MobjSpawn",killwhenland,MT_UNIDUS)

local revertwhenpull = function(mo)
	mo.pulledfunction = function(t,i,table)
		t.state = t.info.spawnstate
		return false
	end
end

addHook("MobjSpawn",revertwhenpull,MT_GOLDBUZZ)
addHook("MobjSpawn",revertwhenpull,MT_REDBUZZ)

addHook("MobjThinker",function(mo)
	if mo.target and mo.target.valid
	and mo.target.state >= S_EGGGUARD_MAD1
	and mo.health
		mo.state = S_NULL
	end
end,MT_EGGSHIELD)

addHook("MobjThinker",function(mo)
	if not (mo.tracer and mo.tracer.valid
	and mo.tracer.health)
	and mo.health
		P_KillMobj(mo)
	end
end,MT_CRUSHCLAW)

addHook("MobjSpawn",function(mo)
	mo.pulledfunction = function(t,i,table)
		P_DamageMobj(t)
		return true
	end
end,MT_GSNAPPER)

addHook("MobjSpawn",function(mo)
	mo.pulledfunction = function(t,i,table)
		P_DamageMobj(t)
		t.state = t.info.spawnstate
		return false
	end
end,MT_CRUSHSTACEAN)

addHook("MobjSpawn",function(mo)
	mo.landfunction = function(t,i,hit)
		P_DamageMobj(t)
		
		for fof in t.subsector.sector.ffloors()
			if fof.flags & FOF_BUSTUP
				EV_CrumbleChain(t.subsector.sector,fof)
			end
		end
		
		return true
	end
	
	mo.hitfunction = function(i,t)
		P_KillMobj(i)
		return true
	end
end,MT_TNTBARREL)

addHook("PlayerCanEnterSpinGaps",function(p)
	if p.mo and p.mo.valid
	and p.mo.skin == "smb2"
	and p.mo.state == S_PLAY_CARRYCROUCH
		return true
	end
end)

addHook("MobjDeath",function(mo)
	mo.deathstilltimer = TICRATE/2
	S_StopMusic(mo.player)
end,MT_PLAYER)

addHook("MusicChange", function(oldname, newname)
	if consoleplayer
		local character = skins[consoleplayer.skin].name
		if character == "smb2"
			if newname == "_clear"
				if mapheaderinfo[gamemap].bonustype >= 1
					return "m2bwin"
				else
					return "m2vict"
				end
			elseif newname == "_1up"
				return "m2life"
			elseif newname == "_gover"
				return "m2over"
			end
		end
	end
end)


A_ExtraLife = function(mo,a,b)
	if ultimate
	or not (mo.target and mo.target.player)
	or mo.target.skin != "smb2"
		super(mo,a,b)
	else
		if mo.type == MT_1UP_ICON and mo.tracer
			// We're using the overlay, so use the overlay 1up sprite (no text)
			mo.sprite = SPR_TV1P
		end
		
		P_GiveCoopLives(mo.target.player,1,false)
		if use1upSound
		or CV_FindVar("1upsound").value == 1
			S_StartSound(nil,sfx_m2life,mo.target.player)
		else
			P_PlayJingleMusic(mo.target.player,"M2LIFE",0,0,JT_1UP)
			mo.target.player.powers[pw_extralife] = TICRATE*3/2
		end
	end
end